/******************************************************************************
 *  Filename:       firmware.c
 *  Date:           04/03/15
 *  File Verison:   2.0
 *
 *  Author:         Alex Hytha
 *  Team:           LFEV'15
 *
 * ****************************************************************************
 *  Architecture:   Enhanced Midrange PIC
 *  Processor:      PIC16LF1827
 *  Compiler:       MPLAB X XC8
 *
 * ****************************************************************************
 *  Description:    Firmware for the slave pic
 *
 * ****************************************************************************
 *  Pin Assignments:
 *                   LEDS:
 *                   RA6= PLED
 *                   RA7= WDT LED (not implemented on the BMS board)
 *                   RB0= Bypass LED
 *
 *                   ANALOG INPUT CHANNELS:
 *                   RA0/AN1=  Current_diff_dis
 *                   RA1/AN0=  Current_diff_chg
 *                   RB2/AN10= Temp
 *                   RB3/AN9=  Volt
 *
 *****************************************************************************/

#include <xc.h>
#include <stdint.h>


/**  CONFIGURATION  ------------------------------------------------------  **/
#pragma config MCLRE = ON       // MCLR/VPP pin function is digital input
#pragma config FOSC  = INTOSC   // Selects the internal oscillator
#pragma config WDTE  = OFF       // WDT controlled by SWDTEN
#pragma config LVP   = OFF      // Low-voltage programming disabled
#pragma config PLLEN = OFF
#pragma config CPD   = OFF      // external EEPROM writing permitted

// command values
#define RESET_BY    0x00
#define SET_BY      0x01
#define PANIC       0x02
#define RELAX       0x03
#define SET_ADDR    0x04
#define SET_BYTIME  0x05

#define default_id   0x04   // default I2C address
#define id_addr      0x00   // slave_id memory location
#define bytime_addr  0x10   // location of the bypass limit

#define MCLR    LATAbits.LATA5  // pin RA5
#define PLED    LATAbits.LATA6  // pin RA6
#define BYPASS  LATBbits.LATB0  // pin RB0
#define SDA1    LATBbits.LATB1  // pin RB1
#define SCL1    LATBbits.LATB4  // pin RB4
#define ICSPCLK LATBbits.LATB6  // pin RB6
#define ICSPDAT LATBbits.LATB7  // pin RB7


/**  VARIABLES  ----------------------------------------------------------  **/

char data_stage;        // the current stage of data collection (for the adc)
uint16_t bypass_count;  // count the number of timer1 cycles until bypass
uint16_t bypass_limit;  // the maximum number of timer0 cycles until bypass occurs
char panic_flag;        // if set, PANIC!
char blink_flag;        // if set, blink the led

char voltage_h; // the cell voltage high bit
char voltage_l; // the cell voltage low bit
char temp_h;    // the cell temp high bit
char temp_l;    // the cell temp low bit

char rxdata;    // data from the I2C bus


/**  FUNCTIONS  ----------------------------------------------------------  **/
void init_ports();  // initializes the I/O ports (including analog input)
void init_vars();   // initialize variables from memory
void init_timer();  // initialize the timers
void init_I2C();    // initialize the I2C bus
void init_ADC();    // initialize the ADC
void init_isr();    // initialize the interrupt handler

void adc_interrupt();       // runs when the ADC finishes a conversion
void i2c_interrupt();       // runs when an I2C message is received
void timer0_interrupt();    // runs when Timer0 rolls over
void timer1_interrupt();    // runs when Timer1 rolls over

void i2c_read();    // read a command byte from the i2c bus
void i2c_write();   // write 4 bytes to the i2c bus

void interrupt isr(void);   // the interrupt routine
void main();                // the main program loop


/*****************************************************************************/
/**  --------------------------  END OF HEADER  --------------------------  **/
/*****************************************************************************/


/** initialization methods  ----------------------------------------------  **/
void init_ports() {
    LATA   = 0x00;  // clear PortA
    LATB   = 0x00;  // clear PortB

    // set the pin output directions to the table in the file descriptor
    TRISA  = 0b00100000; // set PortA
    TRISB  = 0b00011110; // set PortB

    // set up the ADC ports
    ANSELA = 0b00000000; // there are no inputs from AN0-AN4
    ANSELB = 0b00001100; // RB1 and RB2 are analog inputs
}

void init_vars() {
    bypass_count = 0;   // set the bypass counter to 0

    // load the bypass limit from memory
    bypass_limit = (EEPROM_READ(bytime_addr) << 8) + (EEPROM_READ(bytime_addr + 1));
    if( bypass_limit == 0xFF ) {    // if the bypass limit was never set
        bypass_limit = 600;             // default to 600 cycles (~30s)
    }
    
    panic_flag   = 0;   // set panic mode off
    blink_flag   = 1;   // set the initial blink

    data_stage  = 0;    // set the stage to 0
    voltage_h   = 0;    // clear the voltage
    voltage_l   = 0;    // clear the voltage
    temp_h      = 0;    // clear the temperature
    temp_l      = 0;    // clear the temperature
}

void init_timer() {

    // set timer0 to interrupt every 32.8ms
    OPTION_REGbits.TMR0CS = 0;  // set Timer0 clock to the internal clock
    OPTION_REGbits.PSA    = 0;  // assign a prescaler to Timer0
    OPTION_REGbits.PS0    = 1;
    OPTION_REGbits.PS1    = 1;
    OPTION_REGbits.PS2    = 1;  // set the prescaler rate to 1 : 256

}

void init_I2C() {
    SSP1STATbits.SMP = 0;   // enable slew rate control for high speed mode
    SSP1STATbits.CKE = 1;   // enable SMBus specific inputs
    SSP1STATbits.BF  = 0;   // clear the buffer full status bit

    SSP1CON1bits.SSPEN = 1;      // enable the serial port pins
    SSP1CON1bits.CKP   = 1;      // release the clock to allow transmissions
    SSP1CON1bits.SSPM  = 0b0110; // set to slave mode with 7-bit addressing

    SSP1CON2bits.GCEN = 1;  // enable interrupts on a general call (ID 00h)
    SSP1CON2bits.SEN  = 1;  // enable clock stretching

    SSP1CON3bits.BOEN  = 0; // disable buffer overwrites
    SSP1CON3bits.SDAHT = 0; // set the minimum data hold time to 100ns
    SSP1CON3bits.SBCDE = 0; // disable slave bus collision detection
    SSP1CON3bits.AHEN  = 0; // disable address holding
    SSP1CON3bits.DHEN  = 0; // disable data holding

    SSP1MSK = 0xFF;                 // set the address mask to consider all address bits

    char memory_id = EEPROM_READ(id_addr);  // pull the device ID from memory
    if( memory_id == 0xff ) {   // if it has not been set
        SSP1ADD = default_id;       // set the address to the default
    } else {                    // if the ID has already been set
        SSP1ADD = memory_id;        // set the ID to the address in memory
    }

}

void init_ADC() {

    // generate the fixed voltage reference
    FVRCONbits.FVREN = 1;       // enable fixed reference voltage
    FVRCONbits.ADFVR = 0b10;    // set the fixed voltage to 4.096

    // set the voltage references
    ADCON1bits.ADNREF = 0;      // connect VREF- to Vss
    ADCON1bits.ADPREF = 0b11;   // connect VREF+ to FVREF

    ADCON1bits.ADCS = 0b11;     // set ADC clock to dedicated FRC clock
    ADCON1bits.ADFM = 1;        // set the ADC output to be right-justified

}

void init_isr() {
    INTCONbits.TMR0IF = 0;  // clear the timer0 interrupt flag
    PIR1bits.SSP1IF   = 0;  // clear the I2C interrupt flag
    PIR1bits.ADIF     = 0;  // clear the ADC interrupt flag

    INTCONbits.TMR0IE = 1;  // enable timer0 interrupts
    PIE1bits.SSP1IE   = 1;  // enable I2C interrupts
    PIE1bits.ADIE     = 1;  // enable ADC interrupts

    INTCONbits.PEIE = 1;    // enable peripheral interrupts
    INTCONbits.GIE  = 1;    // enable global interrupts
}


/** interrupt routines  --------------------------------------------------  **/
void timer0_interrupt() {
    INTCONbits.TMR0IF = 0;       // clear the timer interrupt flag
    
    // run this if the device is panicking
    if( panic_flag ) {
        PLED = ~PLED;

    // blink the LED if the flag has been set
    } else if( blink_flag ){
        PLED = 1;
        blink_flag = 0;
    } else {
        PLED = 0;
    }

    // only run this function on the 600th interrupt (every 19.66s)
    if(bypass_count == 1000) {
        BYPASS = 1;     // activate the bypass
    } else {
        bypass_count++; // increment the counter
    }

    // start a round of data collection
    data_stage        = 1;       // this is the first stage of data collection
    ADCON0bits.CHS    = 0b01001; // set the ADC channel to the voltage input
    ADCON0bits.ADON   = 1;       // enable the ADC
    ADCON0bits.GO     = 1;       // start an ADC conversion
}

void adc_interrupt() {
    if( data_stage == 1 ) {         // if no data has been gathered yet

        voltage_h = ADRESH; // update the voltage high bit from the ADC result
        voltage_l = ADRESL; // update the voltage low bit from the ADC result

        PIR1bits.ADIF   = 0;        // clear the ADC interrupt flag
        data_stage      = 2;        // start the second stage of data collection
        ADCON0bits.CHS  = 0b01010;  // set the channel to the temperature input
        ADCON0bits.ADON = 1;        // enable the ADC
        ADCON0bits.GO   = 1;        // start the ADC conversion

    } else if( data_stage == 2 ) {  // if the voltage has already been recorded

        temp_h = ADRESH;    // update the temp high bit from the ADC result
        temp_l = ADRESL;    // update the temp low bit from the ADC result

        PIR1bits.ADIF   = 0;         // clear the ADC interrupt flag
        data_stage      = 0;         // start the data collection from the beginning

    }
}

void i2c_interrupt() {
    PIR1bits.SSP1IF = 0;    // clear the interrupt flag
    blink_flag = 1; // indicate that a message is processing

    if(SSP1CON1bits.CKP == 0) { // only respond if the clock line waiting for a response

        if(SSP1STATbits.R_nW == 0)
            i2c_read();     // if the master is requesting data be written
        else
            i2c_write();    // if the master is trying to write data

    }

}


/** subroutines  ---------------------------------------------------------  **/
void i2c_read() {
    rxdata = SSP1BUF;       // read the address from the i2c line
    SSP1CON1bits.CKP = 1;   // release the clock to get the next message

    while(!PIR1bits.SSP1IF) {}  // wait for another I2C message
    PIR1bits.SSP1IF = 0;        // clear the interrupt flag
    rxdata = SSP1BUF;           // read the command from the i2c line

    if( rxdata == RESET_BY ) {   // RESET BYPASS command
        BYPASS = 0;                         // stop bypassing
        bypass_count = 0;                   // reset the bypass counter

    } else if( rxdata == SET_BY ) {     // SET_BY command
        BYPASS = 1;                         // activate bypass mode

    } else if( rxdata == PANIC ) {      // PANIC command
        panic_flag = 1;

    } else if( rxdata == RELAX ) {      // RELAX command
        PLED = 0;
        panic_flag = 0;

    } else if( rxdata == SET_ADDR ) {   // SET_ADDR command

        SSP1CON1bits.CKP = 1;               // release the clock to get the next message
        while(!PIR1bits.SSP1IF) {}          // wait for another I2C message
        PIR1bits.SSP1IF = 0;                // clear the interrupt flag
        rxdata = SSP1BUF;                   // read the command from the i2c line

        EEPROM_WRITE(id_addr, rxdata);      // write the new address to memory
        SSP1ADD = rxdata;                   // initialize the new address

    } else if( rxdata == SET_BYTIME ) { // SET_BYTIME command
        bypass_limit = 0xFFFF;                // set the limit to the max

        SSP1CON1bits.CKP = 1;               // release the clock to get the next message
        while(!PIR1bits.SSP1IF) {}          // wait for another I2C message
        PIR1bits.SSP1IF = 0;                // clear the interrupt flag
        rxdata = SSP1BUF;                   // read the command from the i2c line

        EEPROM_WRITE(bytime_addr, rxdata);  // save the upper byte of the limit

        SSP1CON1bits.CKP = 1;               // release the clock to get the next message
        while(!PIR1bits.SSP1IF) {}          // wait for another I2C message
        PIR1bits.SSP1IF = 0;                // clear the interrupt flag
        rxdata = SSP1BUF;                   // read the command from the i2c line

        EEPROM_WRITE(bytime_addr + 1, rxdata);  // save the lower byte of the limit
        
        bypass_limit = (EEPROM_READ(bytime_addr) << 8) + (EEPROM_READ(bytime_addr + 1));

    }

    SSP1CON1bits.CKP = 1;   // release the I2C clock
}

void i2c_write() {
    rxdata = SSP1BUF;       // read the address from the i2c line

    SSP1BUF = voltage_h;            // write the upper voltage byte
    SSP1CON1bits.CKP = 1;           // release the clock line to send the data

    while(SSP1CON1bits.CKP) {}  // wait for the clock to be held again
    SSP1BUF = voltage_l;            // write the lower voltage byte
    SSP1CON1bits.CKP = 1;           // release the clock line

    while(SSP1CON1bits.CKP) {}  // wait for the clock to be held again
    SSP1BUF = temp_h;               // write the upper temp byte
    SSP1CON1bits.CKP = 1;           // release the clock line

    while(SSP1CON1bits.CKP) {}  // wait for the clock to be held again
    SSP1BUF = temp_l;               // write the lower temp byte
    SSP1CON1bits.CKP = 1;           // release the clock line
}


/** the interrupt routine  -----------------------------------------------  **/
void interrupt isr(void) {

    INTCONbits.PEIE = 0;    // disable peripheral interrupts
    INTCONbits.GIE  = 0;    // disable global interrupts

    if( PIR1bits.SSP1IF ) {     // if the I2C interrupt flag is set
        i2c_interrupt();
    }
    if( INTCONbits.TMR0IF ) {    // if the Timer0 interrupt flag is set
        timer0_interrupt();
    }
    if( PIR1bits.ADIF ) {        // if the ADC interrupt flag is set
        adc_interrupt();
    }

    INTCONbits.PEIE = 1;    // enable peripheral interrupts
    INTCONbits.GIE  = 1;    // enable global interrupts
}


/** the main program loop  -----------------------------------------------  **/
void main(){

    _delay(10);             // delays for 10 cycles
    OSCCON = 0x72;          // switch to fast internal 8 MHz RC Clock

    init_ports();   // initialize the I/O ports
    init_vars();    // initialize the variables
    init_timer();  // initialize the timers
    init_I2C();     // initialize the I2C bus
    init_ADC();     // initialize the ADC
    init_isr();     // initialize the interrupt handler

    while(1){}
}

